import { create } from 'zustand';
import { persist } from 'zustand/middleware';
import { CPU_FUNCTION_MAPING, CPU_FUNCTION_MAPING_END, CPUNAME } from './CPUS';
import { GetResult, Play, PlayerAction, PlayResult } from './game';

export interface SessionInfoPlayResultOver {
    type: 'over',
    winner: PlayResult,
    game: Game;
}

export interface SessionInfo {
    currentUser: string | undefined;
    // Username -> number of games won
    scoreboard: Record<string, number>;
    currentGame: Game | undefined;

    reset: () => void;
    setUser: (user: string | undefined) => void;
    startGame: (cpu: CPUNAME) => void;
    play: (action: PlayerAction) =>
        | SessionInfoPlayResultOver
        | { type: 'play', play: Play };
}

export interface Game {
    cpu: CPUNAME;
    plays: Play[];
}

export const useSession = create<SessionInfo>()(persist<SessionInfo>((set) => ({
    currentUser: undefined,
    scoreboard: {},
    currentGame: undefined,

    reset: () => set({ currentGame: undefined, currentUser: undefined, scoreboard: {} }),
    setUser: (user) => set(() => {
        if (user === undefined) {
            return { currentUser: undefined, currentGame: undefined };
        };
        return { currentUser: user }
    }),
    startGame: (cpu) => set((state) => {
        if (state.currentGame !== undefined) {
            throw new Error("Game already in progress");
        }
        return {
            currentGame: {
                cpu,
                plays: []
            }
        };
    }),
    play: (action) => {
        let result:
            | undefined
            | { type: 'over', winner: 'user' | 'cpu' | 'tie', game: Game }
            | { type: 'play', play: Play }
            = undefined;
        set((state) => {
            if (!state.currentGame || !state.currentUser) throw new Error("Game current not in play");

            const cpuName = state.currentGame.cpu;
            const plays = state.currentGame.plays;
            const cpuFn = CPU_FUNCTION_MAPING[cpuName];
            const cpuPlay = cpuFn(plays);

            const res = GetResult(cpuPlay, action);
            plays.push(res);

            // TODO: maybe make the best of out a configurable
            if (plays.length === 3) {
                CPU_FUNCTION_MAPING_END.forEach((fn) => fn(plays, state.currentUser ?? ''));
                let winner: 'tie' | 'user' | 'cpu' = 'tie';
                let cpu = 0;
                let user = 0;
                for (const play of plays) {
                    if (play.winner === 'cpu') {
                        cpu++;
                    } else if (play.winner === 'user') {
                        user++;
                    }
                }
                if (cpu > user) {
                    winner = 'cpu';
                } else if (user > cpu) {
                    winner = 'user';
                }
                result = {
                    type: 'over',
                    winner,
                    game: { ...state.currentGame },
                }

                const scoreboard = state.scoreboard;

                if (winner === 'user') {
                    scoreboard[state.currentUser] = scoreboard[state.currentUser] ? scoreboard[state.currentUser] + 1 : 1;
                } else if (winner === 'cpu') {
                    scoreboard[cpuName] = scoreboard[cpuName] ? scoreboard[cpuName] + 1 : 1;
                }

                return {
                    currentGame: undefined,
                    scoreboard: { ...scoreboard },
                }
            }

            result = {
                type: 'play',
                play: res,
            };

            return {
                currentGame: { ...state.currentGame, plays: [...plays] }
            };
        })
        if (!result) throw new Error('unrechable');
        return result;
    },
}), {
    name: 'session-info',
}))
